// findwinDlg.cpp : implementation file
//

#include "stdafx.h"
#include "findwin.h"
#include "findwinDlg.h"
#include ".\findwindlg.h"
#include <strsafe.h>
#include <tlhelp32.h>
#include <algorithm>

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

class CWinFinder
{
public:
	CTreeCtrl &m_tvcWindows;
	DWORD m_pid;

protected:
	void Fill(HWND hwnd, HTREEITEM item);

public:
	CWinFinder(DWORD pid, CTreeCtrl &tvc) : m_tvcWindows(tvc), m_pid(pid) {}
	void AddItem(HWND hwnd, HTREEITEM item);
};

void CWinFinder::AddItem(HWND hwnd, HTREEITEM item)
{
	TCHAR	desc[4096];
	TCHAR	windowName[MAX_PATH];
	TCHAR	className[MAX_PATH];

	SendMessage(hwnd, WM_GETTEXT, MAX_PATH, (LPARAM) windowName);

	GetClassName(hwnd, className, sizeof(className));
	StringCbPrintf(desc, sizeof(desc), TEXT("%08X \"%s\" %s"), hwnd, windowName, className);

	BOOL visible = IsWindowVisible(hwnd);

	HTREEITEM newItem = m_tvcWindows.InsertItem(desc, !visible, !visible, item, TVI_LAST);
	m_tvcWindows.SetItemData(newItem, (DWORD_PTR) hwnd);

	for(HWND child = ::GetWindow(hwnd, GW_CHILD);
		child != NULL;
		child = ::GetWindow(child, GW_HWNDNEXT))
	{
		AddItem(child, newItem);
	}
}


BOOL CALLBACK FindWindowByPidProc(HWND hwnd, LPARAM param)
{
	CWinFinder *finder = (CWinFinder *) param;

	DWORD pid;
	GetWindowThreadProcessId(hwnd, &pid);
	if(pid == finder->m_pid)
	{
		finder->AddItem(hwnd, NULL);
	}

	return TRUE;
}

//////////



HWND FindSmallestChildWindowFromPoint(HWND hwnd, CPoint &pt)
{
	HWND child = GetWindow(hwnd, GW_CHILD);
	HWND result = NULL;

	if(child)
	{
		HWND h;
		CRect rc;
		CRect tmp;
		BOOL isSmall;

		// ڽ Ѵ.
		for(child = GetWindow(hwnd, GW_CHILD);
			child != NULL;
			child = GetWindow(child, GW_HWNDNEXT))
		{
			// ڱ⺸    ڽ  ã´.
			h = FindSmallestChildWindowFromPoint(child, pt);

			// ã  Ѵ.
			if(h)
				return h;
			
			GetWindowRect(child, &tmp);

			//    콺 Ͱ Եǰ
			// ó ̰ų  ã    
			isSmall = tmp.Width() <= rc.Width() && tmp.Height() <= rc.Height();
			if(tmp.PtInRect(pt) && (!result || isSmall))
			{
				result = child;
				rc = tmp;
			}
		}
	}

	return result;
}

HWND FindSmallestWindowFromPoint(CPoint &pt)
{
	HWND hwnd = WindowFromPoint(pt);
	if(hwnd)
	{
		HWND child = FindSmallestChildWindowFromPoint(hwnd, pt);
		if(child)
			return child;
	}

	return hwnd;
}


/////////


// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	enum { IDD = IDD_ABOUTBOX };

	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support

// Implementation
protected:
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
END_MESSAGE_MAP()


// CfindwinDlg dialog


CfindwinDlg::CfindwinDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CfindwinDlg::IDD, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDI_FINDER);
	m_finding = FALSE;
	m_watchWnd = NULL;
}

void CfindwinDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_TVC_RESULT, m_tvcResult);
	DDX_Control(pDX, IDC_CMB_PROCESS, m_cmbProcess);
	DDX_Control(pDX, IDC_STC_FINDER, m_stcFinder);
}

BEGIN_MESSAGE_MAP(CfindwinDlg, CDialog)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDC_BTN_FIND, OnBnClickedBtnFind)
	ON_STN_CLICKED(IDC_STC_FINDER, OnStnClickedStcFinder)
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_CAPTURECHANGED()
	ON_WM_SIZE()
	ON_WM_GETMINMAXINFO()
	ON_CBN_SELCHANGE(IDC_CMB_PROCESS, OnCbnSelchangeCmbProcess)
	ON_BN_CLICKED(IDC_BTN_SCANPROCESS, OnBnClickedBtnScanprocess)
	ON_WM_TIMER()
END_MESSAGE_MAP()


BOOL CfindwinDlg::FillProcess()
{
	BOOL ret = FALSE;

	m_cmbProcess.LockWindowUpdate();

	m_pids.clear();
	while(m_cmbProcess.GetCount())
		m_cmbProcess.DeleteString(0);

	HANDLE snap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	if(snap == INVALID_HANDLE_VALUE)
		return TRUE;

	PROCESSENTRY32 pe;
	ZeroMemory(&pe, sizeof(pe));
	pe.dwSize = sizeof(pe);

	if(!Process32First(snap, &pe))
		goto $cleanup;

	do
	{
		m_pids.push_back(pe.th32ProcessID);
		m_cmbProcess.AddString(pe.szExeFile);

	} while(Process32Next(snap, &pe));

	ret = TRUE;

$cleanup:
	CloseHandle(snap);

	m_cmbProcess.UnlockWindowUpdate();
	return ret;
}

BOOL CfindwinDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	////////////////////

	GetWindowRect(&m_initRect);

	CRect client;
	GetClientRect(&client);

	m_tvcResult.GetWindowRect(&m_resultRect);
	ScreenToClient(&m_resultRect);

	m_resultRect.right = client.right - m_resultRect.right;
	m_resultRect.bottom = client.bottom - m_resultRect.bottom;

	m_prevCursor = NULL;
	m_finderCursor = AfxGetApp()->LoadCursor(MAKEINTRESOURCE(IDC_FINDER));

	m_finderIcon = AfxGetApp()->LoadIcon(MAKEINTRESOURCE(IDI_FINDER));
	m_emptyFinderIcon = AfxGetApp()->LoadIcon(MAKEINTRESOURCE(IDI_EMPTYFINDER));

	m_imlType.Create(IDB_BITMAP1, 16, 1, RGB(0,255,0));
/*
	m_imlType.Create(16, 15, ILC_COLORDDB | ILC_MASK, 0, 1);
	CBitmap bmp;
	bmp.LoadBitmap(IDB_BITMAP1);
	m_imlType.Add(&bmp, RGB(0,255,0));
*/
	m_tvcResult.SetImageList(&m_imlType, TVSIL_NORMAL);

	FillProcess();
	m_cmbProcess.SetCurSel(0);

	SetTimer(1, 3000, NULL);
	
	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CfindwinDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CfindwinDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CfindwinDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CfindwinDlg::ChangeFinderIcon()
{
	if(m_finding)
		m_stcFinder.SetIcon(m_emptyFinderIcon);
	else
		m_stcFinder.SetIcon(m_finderIcon);
}


void CfindwinDlg::OnBnClickedBtnFind()
{


}


void CfindwinDlg::StartFindWindow()
{
	m_finding = TRUE;
	m_watchWnd = 0;

	ChangeFinderIcon();
	m_prevCursor = SetCursor(m_finderCursor);
	//ShowWindow(SW_HIDE);
	SetCapture();
}

void CfindwinDlg::StopFindWindow()
{
	m_finding = FALSE;

	ChangeFinderIcon();
	SetCursor(m_prevCursor);

	if(IsWindow(m_watchWnd))
	{
		DWORD pid;
		GetWindowThreadProcessId(m_watchWnd, &pid);

		DwVec::iterator it = std::find(m_pids.begin(), m_pids.end(), pid);
		if(it != m_pids.end())
		{
			int index = int(it - m_pids.begin());
			m_cmbProcess.SetCurSel(index);
            OnCbnSelchangeCmbProcess();
			m_tvcResult.SetFocus();
		}
	}

	//ShowWindow(SW_SHOW);
	ReleaseCapture();
	DrawBorder();

	m_watchWnd = NULL;
}


void CfindwinDlg::Find()
{
	CPoint pt;

	GetCursorPos(&pt);
	HWND hwnd = FindSmallestWindowFromPoint(pt);
	
	DWORD pid;
	GetWindowThreadProcessId(hwnd, &pid);
	if(hwnd && hwnd != m_watchWnd && pid != GetCurrentProcessId())
	{
		DrawBorder();
		m_watchWnd = hwnd;
		DrawBorder();

	}
}

void CfindwinDlg::OnStnClickedStcFinder()
{
	StartFindWindow();
}

void CfindwinDlg::OnLButtonUp(UINT nFlags, CPoint point)
{
	if(m_finding)
		StopFindWindow();

	CDialog::OnLButtonUp(nFlags, point);
}

void CfindwinDlg::OnMouseMove(UINT nFlags, CPoint point)
{
	if(m_finding)
		Find();

	CDialog::OnMouseMove(nFlags, point);
}

void CfindwinDlg::DrawBorder()
{
	if(!m_watchWnd || !IsWindow(m_watchWnd))
		return;

	CRect rc;
	CWnd *p = CWnd::FromHandle(m_watchWnd);

	p->GetWindowRect(&rc);
	CDC *dc = p->GetWindowDC();

	if(dc)
	{
		CBrush brs;
		CBrush *old;
		CPen pen;
		int prevMode;

		pen.CreatePen(PS_INSIDEFRAME, 4, RGB(255,255,255));
		brs.CreateStockObject(HOLLOW_BRUSH);
		
		prevMode = dc->SetROP2(R2_XORPEN);
		old = (CBrush *) dc->SelectObject(&brs);
		dc->SelectObject(&pen);

		if(p->GetStyle() & WS_MAXIMIZE)
			dc->Rectangle(-rc.left, -rc.top, rc.Width() + rc.left, rc.Height() + rc.top);
		else
			dc->Rectangle(0, 0, rc.Width(), rc.Height());
		dc->SelectObject(old);
		dc->SetROP2(prevMode);

		p->ReleaseDC(dc);
	}
}

void CfindwinDlg::OnCaptureChanged(CWnd *pWnd)
{
	if(m_finding)
		StopFindWindow();

	CDialog::OnCaptureChanged(pWnd);
}

void CfindwinDlg::OnSize(UINT nType, int cx, int cy)
{
	CDialog::OnSize(nType, cx, cy);

	if(m_tvcResult.GetSafeHwnd())
	{
		CRect rc;
		GetClientRect(&rc);

		rc.top += m_resultRect.top;
		rc.left += m_resultRect.left;
		rc.right -= m_resultRect.right;
		rc.bottom -= m_resultRect.bottom;

		m_tvcResult.MoveWindow(&rc);
	}
}

void CfindwinDlg::OnGetMinMaxInfo(MINMAXINFO* lpMMI)
{
	if(m_tvcResult.GetSafeHwnd())
	{
		lpMMI->ptMinTrackSize.x = m_initRect.Width();
		lpMMI->ptMinTrackSize.y = m_initRect.Height();
	}

	CDialog::OnGetMinMaxInfo(lpMMI);
}

void CfindwinDlg::OnCbnSelchangeCmbProcess()
{
	int sel = m_cmbProcess.GetCurSel();
	if(sel < 0)
		return;

	m_tvcResult.LockWindowUpdate();
	m_tvcResult.DeleteAllItems();

	DWORD pid = m_pids[sel];
	CWinFinder finder(pid, m_tvcResult);
	EnumWindows(FindWindowByPidProc, (LPARAM) &finder);
	m_tvcResult.UnlockWindowUpdate();
}

void CfindwinDlg::OnBnClickedBtnScanprocess()
{
	int sel = m_cmbProcess.GetCurSel();
	DWORD pid = m_pids[sel];

	FillProcess();

	DwVec::iterator it = std::find(m_pids.begin(), m_pids.end(), pid);
	if(it != m_pids.end())
		m_cmbProcess.SetCurSel(int(it - m_pids.begin()));
}

void CfindwinDlg::OnTimer(UINT nIDEvent)
{
	if(nIDEvent == 1)
	{
		OnBnClickedBtnScanprocess();
		
	}

	CDialog::OnTimer(nIDEvent);
}
